/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2009 Novell, Inc.
 */

#include <glib/gi18n.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <moblin-panel/mpl-panel-gtk.h>
#include <nbtk/nbtk-gtk.h>
#include "nmn-applet.h"
#include "nmn-nm-data.h"
#include "nmn-status-icon.h"
#include "nmn-networks.h"
#include "nmn-new-connection.h"

G_DEFINE_TYPE (NmnApplet, nmn_applet, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_PANEL_CLIENT,
    PROP_WINDOW,

    LAST_PROP
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NMN_TYPE_APPLET, NmnAppletPrivate))

typedef struct {
    NmnNMData *nm_data;
    MplPanelClient *panel_client;
    GtkWindow *window;
    NmnStatusIcon *status_icon;

    GtkWidget *pane;
    GtkWidget *list;
    GtkWidget *enable_wifi;
    GtkWidget *enable_ethernet;
    GtkWidget *enable_3g;
    GtkWidget *enable_wimax;
    GtkWidget *enable_bt;
    GtkWidget *enable_network;
    GtkWidget *add_new_connection;

    GtkWidget *new_dialog;

    gboolean network_list_populated;

    gboolean disposed;
} NmnAppletPrivate;

NmnApplet *
nmn_applet_new (MplPanelClient *panel_client)
{
    g_return_val_if_fail (MPL_IS_PANEL_CLIENT (panel_client), NULL);

    return NMN_APPLET (g_object_new (NMN_TYPE_APPLET,
                                     NMN_APPLET_PANEL_CLIENT, panel_client,
                                     NULL));
}

NmnApplet *
nmn_applet_new_standalone (GtkWindow *window)
{
    g_return_val_if_fail (GTK_IS_WINDOW (window), NULL);

    return NMN_APPLET (g_object_new (NMN_TYPE_APPLET,
    				     NMN_APPLET_WINDOW, window,
                                     NULL));
}

/* enable/disable wifi button */

static void
enable_wifi_toggled (NbtkGtkLightSwitch *w,
                     gboolean active,
                     gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    nmn_nm_data_wifi_toggled (priv->nm_data, active);
}

static void
wifi_toggle_set_sensitive (NmnNMData *nm_data,
                           GtkWidget *widget,
                           gboolean suggested)
{
    if (suggested == FALSE || nmn_nm_data_wifi_can_change (nm_data) == FALSE)
        gtk_widget_set_sensitive (widget, FALSE);
    else
        gtk_widget_set_sensitive (widget, TRUE);
}

static void
wifi_toggled (NmnNMData *nm_data,
              gboolean active,
              gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    g_signal_handlers_block_by_func (priv->nm_data, enable_wifi_toggled, user_data);
    nbtk_gtk_light_switch_set_active (NBTK_GTK_LIGHT_SWITCH (priv->enable_wifi), active);
    g_signal_handlers_unblock_by_func (priv->nm_data, enable_wifi_toggled, user_data);

    wifi_toggle_set_sensitive (priv->nm_data, priv->enable_wifi, TRUE);
}

static void
enable_wifi_setup (NmnApplet *applet)
{
    NmnAppletPrivate *priv = GET_PRIVATE (applet);

    g_signal_connect (priv->enable_wifi, "switch-flipped", G_CALLBACK (enable_wifi_toggled), applet);
    gtk_widget_show (priv->enable_wifi);

    g_signal_connect (priv->nm_data, "wifi-toggled", G_CALLBACK (wifi_toggled), applet);
    wifi_toggled (priv->nm_data, nmn_nm_data_wifi_get_active (priv->nm_data), applet);
}

/* enable/disable ethernet button */

static void
enable_ethernet_toggled (NbtkGtkLightSwitch *w,
                         gboolean active,
                         gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    nmn_nm_data_ethernet_toggled (priv->nm_data, active);
}

static void
ethernet_toggled (NmnNMData *nm_data,
                  gboolean active,
                  gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    g_signal_handlers_block_by_func (priv->nm_data, enable_ethernet_toggled, user_data);
    nbtk_gtk_light_switch_set_active (NBTK_GTK_LIGHT_SWITCH (priv->enable_ethernet), active);
    g_signal_handlers_unblock_by_func (priv->nm_data, enable_ethernet_toggled, user_data);
}

static void
enable_ethernet_setup (NmnApplet *applet)
{
    NmnAppletPrivate *priv = GET_PRIVATE (applet);

    g_signal_connect (priv->enable_ethernet, "switch-flipped", G_CALLBACK (enable_ethernet_toggled), applet);
    gtk_widget_show (priv->enable_ethernet);

    g_signal_connect (priv->nm_data, "ethernet-toggled", G_CALLBACK (ethernet_toggled), applet);
    ethernet_toggled (priv->nm_data, nmn_nm_data_ethernet_get_active (priv->nm_data), applet);
}

/* enable/disable 3G button */

static void
enable_3g_toggled (NbtkGtkLightSwitch *w,
                   gboolean active,
                   gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    nmn_nm_data_modems_toggled (priv->nm_data, active);
}

static void
modems_toggled (NmnNMData *nm_data,
                gboolean active,
                gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    g_signal_handlers_block_by_func (priv->nm_data, enable_3g_toggled, user_data);
    nbtk_gtk_light_switch_set_active (NBTK_GTK_LIGHT_SWITCH (priv->enable_3g), active);
    g_signal_handlers_unblock_by_func (priv->nm_data, enable_3g_toggled, user_data);
}

static void
enable_3g_setup (NmnApplet *applet)
{
    NmnAppletPrivate *priv = GET_PRIVATE (applet);

    g_signal_connect (priv->enable_3g, "switch-flipped", G_CALLBACK (enable_3g_toggled), applet);
    gtk_widget_show (priv->enable_3g);

    g_signal_connect (priv->nm_data, "modems-toggled", G_CALLBACK (modems_toggled), applet);
    modems_toggled (priv->nm_data, nmn_nm_data_modems_get_active (priv->nm_data), applet);
}

/* enable/disable Offline mode button */

static void
enable_network_toggled (NbtkGtkLightSwitch *w,
                        gboolean active,
                        gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    gtk_widget_set_sensitive (priv->enable_ethernet, !active);
    gtk_widget_set_sensitive (priv->enable_3g, !active);
    wifi_toggle_set_sensitive (priv->nm_data, priv->enable_wifi, !active);

    nmn_nm_data_offline_mode_toggled (priv->nm_data, active);
}

static void
offline_toggled (NmnNMData *nm_data,
                 gboolean active,
                 gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    g_signal_handlers_block_by_func (priv->nm_data, enable_network_toggled, user_data);
    nbtk_gtk_light_switch_set_active (NBTK_GTK_LIGHT_SWITCH (priv->enable_network), active);
    g_signal_handlers_unblock_by_func (priv->nm_data, enable_network_toggled, user_data);

    gtk_widget_set_sensitive (priv->enable_3g, !active);
    gtk_widget_set_sensitive (priv->enable_ethernet, !active);
    wifi_toggle_set_sensitive (priv->nm_data, priv->enable_wifi, !active);
}

static void
enable_network_setup (NmnApplet *applet)
{
    NmnAppletPrivate *priv = GET_PRIVATE (applet);

    g_signal_connect (priv->enable_network, "switch-flipped", G_CALLBACK (enable_network_toggled), applet);
    gtk_widget_show (priv->enable_network);

    g_signal_connect (priv->nm_data, "offline-mode-toggled", G_CALLBACK (offline_toggled), applet);
    offline_toggled (priv->nm_data, nmn_nm_data_offline_mode_get_active (priv->nm_data), applet);
}

/* add new connection button */

static void
add_new_connection_show (GtkButton *button,
                         gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    gtk_widget_hide (priv->pane);
    gtk_widget_show (priv->new_dialog);
}

static void
add_new_connection_hide (GtkWidget *widget,
                         gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    gtk_widget_hide (priv->new_dialog);
    gtk_widget_show (priv->pane);
}

static void
add_new_connection_setup (NmnApplet *applet)
{
    NmnAppletPrivate *priv = GET_PRIVATE (applet);

    priv->new_dialog = nmn_new_connection_create (priv->nm_data);
    g_object_ref_sink (priv->new_dialog);
    gtk_widget_modify_bg (priv->new_dialog, GTK_STATE_NORMAL, &gtk_widget_get_style (priv->new_dialog)->white);
    g_signal_connect (priv->new_dialog, "hide", G_CALLBACK (add_new_connection_hide), applet);

    g_signal_connect (priv->add_new_connection, "clicked", G_CALLBACK (add_new_connection_show), applet);
}

static void
show_begin_cb (MplPanelClient *panel_client,
               gpointer user_data)
{
    NmnAppletPrivate *priv = GET_PRIVATE (user_data);

    if (!priv->network_list_populated) {
        nmn_networks_populate (NMN_NETWORKS (priv->list));
        priv->network_list_populated = TRUE;
    }
}

static void
hide_end_cb (MplPanelClient *panel_client,
             gpointer user_data)
{
    add_new_connection_hide (NULL, user_data);
}

static DBusGConnection *
init_dbus (void)
{
    DBusGConnection *bus;
    DBusGProxy *proxy;
    GError *error = NULL;
    int request_name_result;

    dbus_connection_set_change_sigpipe (TRUE);

    bus = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
    if (!bus) {
        g_warning ("Could not get the system bus.  Make sure "
                   "the message bus daemon is running!  Message: %s",
                   error->message);
        g_error_free (error);
        return NULL;
    }

    dbus_connection_set_exit_on_disconnect (dbus_g_connection_get_connection (bus), FALSE);

    proxy = dbus_g_proxy_new_for_name (bus,
                                       "org.freedesktop.DBus",
                                       "/org/freedesktop/DBus",
                                       "org.freedesktop.DBus");

    if (!dbus_g_proxy_call (proxy, "RequestName", &error,
                            G_TYPE_STRING, NM_DBUS_SERVICE_USER_SETTINGS,
                            G_TYPE_UINT, DBUS_NAME_FLAG_DO_NOT_QUEUE,
                            G_TYPE_INVALID,
                            G_TYPE_UINT, &request_name_result,
                            G_TYPE_INVALID)) {
        g_warning ("Could not acquire the NetworkManagerUserSettings service.\n"
                   "  Message: '%s'", error->message);
        g_error_free (error);
        goto err;
    }

    if (request_name_result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
        g_warning ("Could not acquire the NetworkManagerUserSettings service "
                   "as it is already taken.  Return: %d",
                   request_name_result);
        goto err;
    }

    return bus;

 err:
    dbus_g_connection_unref (bus);
    return NULL;
}

static void
nmn_applet_init (NmnApplet *applet)
{
    NmnAppletPrivate *priv = GET_PRIVATE (applet);
    DBusGConnection *bus;
    GtkWidget *w;
    GtkWidget *vbox;
    GtkWidget *frame;
    GtkWidget *table;
    char *label;

    bus = init_dbus ();
    /* FIXME: Do something nicer here */
    g_assert (bus);

    priv->nm_data = nmn_nm_data_new (bus);
    dbus_g_connection_unref (bus);

    priv->status_icon = nmn_status_icon_new ();

    dbus_g_connection_register_g_object (nm_object_get_connection (NM_OBJECT (priv->nm_data)),
	                                     NM_DBUS_PATH_SETTINGS,
	                                     G_OBJECT (nmn_nm_data_get_user_settings (priv->nm_data)));

    nmn_status_icon_set_nm_client (priv->status_icon, NM_CLIENT (priv->nm_data));

    priv->pane = gtk_hbox_new (FALSE, 6);
    gtk_container_set_border_width (GTK_CONTAINER (priv->pane), 6);

    /* left side (Networks list, add new connection button) */
    vbox = gtk_vbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX (priv->pane), vbox, TRUE, TRUE, 0);

    frame = nbtk_gtk_frame_new ();
    gtk_box_pack_start (GTK_BOX (vbox), frame, TRUE, TRUE, 0);

    priv->add_new_connection = gtk_button_new_with_label (_("Add new connection"));
    gtk_box_pack_start (GTK_BOX (vbox), priv->add_new_connection, FALSE, FALSE, 0);

    vbox = gtk_vbox_new (FALSE, 6);
    gtk_container_add (GTK_CONTAINER (frame), vbox);

    label = g_strdup_printf ("<big><b>%s</b></big>", _("Networks"));
    w = gtk_label_new (NULL);
    gtk_label_set_markup (GTK_LABEL (w), label);
    g_free (label);
    gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
    gtk_box_pack_start (GTK_BOX (vbox), w, FALSE, FALSE, 0);

    w = gtk_scrolled_window_new (NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (w), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    priv->list = nmn_networks_new (priv->nm_data);
    gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (w), priv->list);
    gtk_widget_show (priv->list);
    gtk_box_pack_start (GTK_BOX (vbox), w, TRUE, TRUE, 0);


    /* Right side, switches */
    vbox = gtk_vbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX (priv->pane), vbox, FALSE, FALSE, 0);

    frame = nbtk_gtk_frame_new ();
    gtk_box_pack_start (GTK_BOX (vbox), frame, TRUE, TRUE, 0);

    table = gtk_table_new (5, 2, TRUE);
    gtk_table_set_row_spacings (GTK_TABLE (table), 6);
    gtk_container_set_border_width (GTK_CONTAINER (table), 6);
    gtk_container_add (GTK_CONTAINER (frame), table);

    w = gtk_label_new (_("WiFi"));
    gtk_misc_set_alignment (GTK_MISC (w), 0.2, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 1, 0, 1);
    priv->enable_wifi = nbtk_gtk_light_switch_new ();
    gtk_table_attach_defaults (GTK_TABLE (table), priv->enable_wifi, 1, 2, 0, 1);

    w = gtk_label_new (_("Wired"));
    gtk_misc_set_alignment (GTK_MISC (w), 0.2, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 1, 1, 2);
    priv->enable_ethernet = nbtk_gtk_light_switch_new ();
    gtk_table_attach_defaults (GTK_TABLE (table), priv->enable_ethernet, 1, 2, 1, 2);

    w = gtk_label_new (_("3G"));
    gtk_misc_set_alignment (GTK_MISC (w), 0.2, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 1, 2, 3);
    priv->enable_3g = nbtk_gtk_light_switch_new ();
    gtk_table_attach_defaults (GTK_TABLE (table), priv->enable_3g, 1, 2, 2, 3);

    w = gtk_label_new (_("WiMAX"));
    gtk_misc_set_alignment (GTK_MISC (w), 0.2, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 1, 3, 4);
    priv->enable_wimax = nbtk_gtk_light_switch_new ();
    gtk_widget_set_sensitive (priv->enable_wimax, FALSE);
    gtk_table_attach_defaults (GTK_TABLE (table), priv->enable_wimax, 1, 2, 3, 4);

    w = gtk_label_new (_("Bluetooth"));
    gtk_misc_set_alignment (GTK_MISC (w), 0.2, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 1, 4, 5);
    priv->enable_bt = nbtk_gtk_light_switch_new ();
    gtk_widget_set_sensitive (priv->enable_bt, FALSE);
    gtk_table_attach_defaults (GTK_TABLE (table), priv->enable_bt, 1, 2, 4, 5);


    frame = nbtk_gtk_frame_new ();
    gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);

    table = gtk_table_new (2, 2, TRUE);
    gtk_container_add (GTK_CONTAINER (frame), table);

    w = gtk_label_new (_("Offline mode"));
    gtk_misc_set_alignment (GTK_MISC (w), 0.2, 0.5);
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 1, 0, 1);
    priv->enable_network = nbtk_gtk_light_switch_new ();
    gtk_table_attach_defaults (GTK_TABLE (table), priv->enable_network, 1, 2, 0, 1);

    w = gtk_label_new (_("This will disable all your connections"));
    gtk_table_attach_defaults (GTK_TABLE (table), w, 0, 2, 1, 2);

    gtk_widget_show_all (priv->pane);


    enable_wifi_setup (applet);
    enable_ethernet_setup (applet);
    enable_3g_setup (applet);
    enable_network_setup (applet);
    add_new_connection_setup (applet);
}

static GObject*
constructor (GType type,
             guint n_construct_params,
             GObjectConstructParam *construct_params)
{
    GObject *object;
    NmnAppletPrivate *priv;
    GtkWidget *box;
    GtkWidget *window;

    object = G_OBJECT_CLASS (nmn_applet_parent_class)->constructor
        (type, n_construct_params, construct_params);

    if (!object)
        return NULL;

    priv = GET_PRIVATE (object);

    if (!priv->panel_client && !priv->window) {
        g_warning ("Missing constructor arguments");
        g_object_unref (object);
        return NULL;
    }
    if (priv->panel_client && priv->window) {
        g_warning ("You can only use a PanelClient if you don't pass a window");
        g_object_unref (object);
        return NULL;
    }

    box = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (priv->pane);
    gtk_box_pack_start (GTK_BOX (box), priv->pane, TRUE, TRUE, 0);

    gtk_widget_hide (priv->new_dialog);
    gtk_box_pack_end (GTK_BOX (box), priv->new_dialog, TRUE, TRUE, 0);

    gtk_widget_show (box);
    if (priv->panel_client)
        window = mpl_panel_gtk_get_window (MPL_PANEL_GTK (priv->panel_client));
    else
        window = GTK_WIDGET (priv->window);
    gtk_container_add (GTK_CONTAINER (window), box);
    gtk_widget_modify_bg (window, GTK_STATE_NORMAL, &gtk_widget_get_style (window)->white);
    gtk_widget_show (GTK_WIDGET (window));

    if (priv->panel_client) {
        g_signal_connect (priv->panel_client, "show-begin", G_CALLBACK (show_begin_cb), object);
        g_signal_connect (priv->panel_client, "hide-end", G_CALLBACK (hide_end_cb), object);
    } else {
        show_begin_cb (NULL, object);
    }

    return object;
}

static void
set_property (GObject *object, guint prop_id,
              const GValue *value, GParamSpec *pspec)
{
    NmnAppletPrivate *priv = GET_PRIVATE (object);

    switch (prop_id) {
    case PROP_PANEL_CLIENT:
        /* Construct only */
        priv->panel_client = g_value_dup_object (value);
        if (priv->panel_client)
            nmn_status_icon_set_panel_client (priv->status_icon, priv->panel_client);
        break;
    case PROP_WINDOW:
    	/* Construct only */
        priv->window = g_value_dup_object (value);
        //FIXME we'd like to show a status icon somewhere though
#if 0
        if (priv->window)
            nmn_status_icon_set_panel_client (priv->status_icon, priv->panel_client);
#endif
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
get_property (GObject *object, guint prop_id,
              GValue *value, GParamSpec *pspec)
{
    NmnAppletPrivate *priv = GET_PRIVATE (object);

    switch (prop_id) {
    case PROP_PANEL_CLIENT:
        g_value_set_object (value, priv->panel_client);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
dispose (GObject *object)
{
    NmnAppletPrivate *priv = GET_PRIVATE (object);

    if (priv->disposed)
        return;

    priv->disposed = TRUE;

    if (priv->new_dialog)
        gtk_widget_destroy (priv->new_dialog);

    if (priv->pane)
        g_object_unref (priv->pane);

    if (priv->status_icon)
        g_object_unref (priv->status_icon);

    if (priv->nm_data)
        g_object_unref (priv->nm_data);

    if (priv->panel_client)
        g_object_unref (priv->panel_client);

    if (priv->window)
        g_object_unref (priv->window);

    G_OBJECT_CLASS (nmn_applet_parent_class)->dispose (object);
}

static void
nmn_applet_class_init (NmnAppletClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);

    g_type_class_add_private (object_class, sizeof (NmnAppletPrivate));

    object_class->constructor = constructor;
    object_class->set_property = set_property;
    object_class->get_property = get_property;
    object_class->dispose = dispose;

    /* properties */
    g_object_class_install_property
        (object_class, PROP_PANEL_CLIENT,
         g_param_spec_object (NMN_APPLET_PANEL_CLIENT,
                              "MplPanelClient",
                              "Panel client",
                              MPL_TYPE_PANEL_CLIENT,
                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property
        (object_class, PROP_WINDOW,
         g_param_spec_object (NMN_APPLET_WINDOW,
                              "GtkWindow",
                              "Top-level window",
                              GTK_TYPE_WINDOW,
                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
}
