/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2009 Novell, Inc.
 */

#include <glib/gi18n.h>
#include <NetworkManager.h>
#include <nm-device.h>
#include <nm-connection.h>
#include <nm-setting-connection.h>
#include <nm-setting-ip4-config.h>
#include "nmn-network-item.h"
#include "nmn-icon-cache.h"
#include "nma-gconf-connection.h"

#define DBUS_TYPE_G_MAP_OF_VARIANT (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))
#define DBUS_TYPE_G_MAP_OF_MAP_OF_VARIANT (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, DBUS_TYPE_G_MAP_OF_VARIANT))

static void item_init (NmnItem *item_class);

G_DEFINE_TYPE_EXTENDED (NmnNetworkItem, nmn_network_item, GTK_TYPE_EVENT_BOX, 0,
                        G_IMPLEMENT_INTERFACE (NMN_TYPE_ITEM, item_init))

enum {
    PROP_0,
    PROP_NM_DATA,
    PROP_DEVICE,
    PROP_CONNECTION,
    PROP_AC,
    PROP_TYPE_STRING,

    LAST_PROP
};

enum {
    CONNECT_REQUESTED,
    DISCONNECT_REQUESTED,
    DELETE,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NMN_TYPE_NETWORK_ITEM, NmnNetworkItemPrivate))

typedef struct {
    GtkBox *vbox; /* child of self */
    GtkBox *hbox; /* child of vbox */

    GtkWidget *icon;
    GtkLabel *name;
    GtkWidget *status_separator;
    GtkLabel *status_label;
    GtkLabel *security;
    GtkWidget *expander;
    NmnConnectionDetails *details;
    GtkButton *connect;
    GtkWidget *remove;
    NmnNetworkItemStatus status;

    NmnNMData *nm_data;
    NMDevice *device;
    NMSettingsConnectionInterface *connection;
    NMActiveConnection *ac;
    char *type_str;

    GdkColor prelight_color;

    gulong connection_updated_id;
    gulong connection_removed_id;
    gulong secrets_requested_id;
    gulong ac_state_changed_id;

    gboolean disposed;
} NmnNetworkItemPrivate;

NmnNMData *
nmn_network_item_get_nm_data (NmnNetworkItem *self)
{
    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (self), NULL);

    return GET_PRIVATE (self)->nm_data;
}

NMDevice *
nmn_network_item_get_device (NmnNetworkItem *self)
{
    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (self), NULL);

    return GET_PRIVATE (self)->device;
}

NMActiveConnection *
nmn_network_item_get_active_connection (NmnNetworkItem *self)
{
    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (self), NULL);

    return GET_PRIVATE (self)->ac;
}

static void
update_details (NmnNetworkItem *self)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (self);
    NMConnection *connection;
    NMSettingIP4Config *setting;
    NMIP4Config *config;

    if (!priv->details || !priv->connection)
        return;

    connection = NM_CONNECTION (priv->connection);
    setting = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);

    if (nm_device_get_state (priv->device) == NM_DEVICE_STATE_ACTIVATED)
        config = nm_device_get_ip4_config (priv->device);
    else
        config = NULL;

    nmn_connection_details_set_data (priv->details, setting, config);
}

static void
ac_state_changed (NMActiveConnection *ac,
                  GParamSpec *pspec,
                  gpointer user_data)
{
    NmnNetworkItem *item = NMN_NETWORK_ITEM (user_data);
    NmnNetworkItemStatus status;

    if (ac) {
        switch (nm_active_connection_get_state (ac)) {
        case NM_ACTIVE_CONNECTION_STATE_ACTIVATED:
            status = NMN_NETWORK_ITEM_STATUS_CONNECTED;
            break;
        case NM_ACTIVE_CONNECTION_STATE_ACTIVATING:
            status = NMN_NETWORK_ITEM_STATUS_CONNECTING;
            break;
        default:
            status = NMN_NETWORK_ITEM_STATUS_DISCONNECTED;
            break;
        }
    } else
        status = NMN_NETWORK_ITEM_STATUS_DISCONNECTED;

    nmn_network_item_set_status (item, status);
    g_object_notify (G_OBJECT (item), NMN_ITEM_PRIORITY);
    update_details (item);
}

void
nmn_network_item_set_active_connection (NmnNetworkItem *self,
                                        NMActiveConnection *ac)
{
    NmnNetworkItemPrivate *priv;

    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    priv = GET_PRIVATE (self);
    if (priv->ac) {
        g_signal_handler_disconnect (priv->ac, priv->ac_state_changed_id);
        g_object_unref (priv->ac);
    }

    if (ac) {
        priv->ac = g_object_ref (ac);
        priv->ac_state_changed_id = g_signal_connect (ac, "notify::" NM_ACTIVE_CONNECTION_STATE,
                                                      G_CALLBACK (ac_state_changed),
                                                      self);
    } else {
        priv->ac = NULL;
        priv->ac_state_changed_id = 0;
    }

    ac_state_changed (ac, NULL, self);
}

static void
update_item (NmnNetworkItem *self)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (self);
    NmnItem *item = NMN_ITEM (self);
    NMConnection *wrapped;
    NMSettingConnection *s_con;

    g_assert (priv->connection);

    wrapped = NM_CONNECTION (priv->connection);
    s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (wrapped, NM_TYPE_SETTING_CONNECTION));

    nmn_item_set_name (item, nm_setting_connection_get_id (s_con));
}

static void
updated (NMExportedConnection *connection,
         GHashTable *settings,
         gpointer user_data)
{
    update_item (NMN_NETWORK_ITEM (user_data));
}

static void
removed (NMExportedConnection *connection, gpointer user_data)
{
    nmn_item_remove_requested (NMN_ITEM (user_data));
}


void
nmn_network_item_set_status_visible (NmnNetworkItem *self,
                                     gboolean visible)
{
    NmnNetworkItemPrivate *priv;

    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    priv = GET_PRIVATE (self);
    if (visible) {
        gtk_widget_show (priv->status_separator);
        gtk_widget_show (GTK_WIDGET (priv->status_label));
    } else {
        gtk_widget_hide (priv->status_separator);
        gtk_widget_hide (GTK_WIDGET (priv->status_label));
    }
}

void
nmn_network_item_set_status (NmnNetworkItem *self,
                             NmnNetworkItemStatus status)
{
    NmnNetworkItemPrivate *priv;

    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    priv = GET_PRIVATE (self);
    if (priv->status != status) {
        const char *status_msg;
        char *str;

        priv->status = status;

        switch (status) {
        case NMN_NETWORK_ITEM_STATUS_DISCONNECTED:
            status_msg = _("Disconnected");
            gtk_button_set_label (priv->connect, _("Connect"));
            break;
        case NMN_NETWORK_ITEM_STATUS_CONNECTING:
            status_msg = _("Connecting...");
            gtk_button_set_label (priv->connect, _("Disconnect"));
            break;
        case NMN_NETWORK_ITEM_STATUS_CONNECTED:
            status_msg = _("Connected");
            gtk_button_set_label (priv->connect, _("Disconnect"));
            break;
        default:
            g_assert_not_reached ();
        }

        str = g_strconcat ("<big><b>", status_msg, "</b></big>", NULL);
        gtk_label_set_markup (priv->status_label, str);
        g_free (str);

        nmn_network_item_set_status_visible (self, TRUE);
    }
}

NmnNetworkItemStatus
nmn_network_item_get_status (NmnNetworkItem *self)
{
    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (self), NMN_NETWORK_ITEM_STATUS_DISCONNECTED);

    return GET_PRIVATE (self)->status;
}

void
nmn_network_item_set_icon (NmnNetworkItem *self,
                           const char *icon_name)
{
    GtkImage *image;
    GdkPixbuf *pixbuf;

    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));
    g_return_if_fail (icon_name != NULL);

    pixbuf = nmn_icon_cache_get (icon_name);
    if (!pixbuf)
        return;

    image = GTK_IMAGE (GET_PRIVATE (self)->icon);

    if (gtk_image_get_storage_type (image) == GTK_IMAGE_PIXBUF && gtk_image_get_pixbuf (image) == pixbuf)
        return;

    gtk_image_set_from_pixbuf (image, pixbuf);
}

static void
connection_secrets_requested_cb (NMAGConfConnection *connection,
                                 const char *setting_name,
                                 const char **hints,
                                 gboolean ask_user,
                                 DBusGMethodInvocation *context,
                                 gpointer user_data)
{
    NmnNetworkItem *self = NMN_NETWORK_ITEM (user_data);

    if (NMN_NETWORK_ITEM_GET_CLASS (self)->secrets_requested)
        NMN_NETWORK_ITEM_GET_CLASS (self)->secrets_requested (self, setting_name, hints, ask_user, context);
    else {
        GError *error = NULL;

        g_set_error_literal (&error, NM_SETTINGS_INTERFACE_ERROR, NM_SETTINGS_INTERFACE_ERROR_SECRETS_UNAVAILABLE,
                             G_STRLOC "no secrets found");

        g_warning ("%s", error->message);
        dbus_g_method_return_error (context, error);
        g_error_free (error);
    }
}

static void
connect_cb (gpointer user_data,
            const char *object_path,
            GError *error)
{
    /* FIXME: Report the error somewhere */
}

static void
update_cb (NMSettingsConnectionInterface *connection,
	   GError *error,
	   gpointer data)
{
    NmnItem *item = NMN_ITEM (data);
    NmnNetworkItemPrivate *priv = GET_PRIVATE (item);
    NMConnection *wrapped;
    const char *path;
    const char *service_name;
    const char *specific_object;
    NMConnectionScope scope;

    if (error != NULL) {
    	    g_warning ("Updating settings failed: %s", error->message);
    	    return;
    }

    wrapped = NM_CONNECTION (priv->connection);
    path = nm_connection_get_path (wrapped);
    scope = nm_connection_get_scope (wrapped);

    service_name = (scope == NM_CONNECTION_SCOPE_USER) ?
        NM_DBUS_SERVICE_USER_SETTINGS : NM_DBUS_SERVICE_SYSTEM_SETTINGS;

    if (NMN_NETWORK_ITEM_GET_CLASS (item)->get_specific_object)
        specific_object = NMN_NETWORK_ITEM_GET_CLASS (item)->get_specific_object (NMN_NETWORK_ITEM (item));
    else
        specific_object = "/";

    nm_client_activate_connection (NM_CLIENT (priv->nm_data),
                                   service_name,
                                   path,
                                   priv->device,
                                   specific_object,
                                   connect_cb,
                                   NULL);
}

static gboolean
delayed_connect (gpointer data)
{
	update_cb (NULL, NULL, data);
	return FALSE;
}

static void
connect (NmnNetworkItem *item)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (item);

    if (!priv->connection) {
        g_warning ("Connection not set");
        return;
    }

    if (nmn_network_item_has_details (item)) {
        NmnConnectionDetails *details = nmn_network_item_get_details (item);
        NMConnection *wrapped;
        NMSetting *current_config;
        NMSetting *new_config;

        wrapped = NM_CONNECTION (priv->connection);
        current_config = nm_connection_get_setting (wrapped, NM_TYPE_SETTING_IP4_CONFIG);
        new_config = NM_SETTING (nmn_connection_details_get_data (NMN_CONNECTION_DETAILS (details)));
        g_assert (new_config);

        if (current_config == NULL || nm_setting_compare (current_config, new_config, 0) == FALSE) {
            nm_connection_add_setting (wrapped, new_config);
            nm_settings_connection_interface_update (NM_SETTINGS_CONNECTION_INTERFACE (new_config),
            					     update_cb, item);

            return;
        }
    }

    g_idle_add (delayed_connect, item);
}

static void
disconnect_cb (NMDevice *device, GError *error, gpointer user_data)
{
    if (error)
        g_warning ("Device disconnect failed: (%d) %s",
                   error ? error->code : -1,
                   error && error->message ? error->message : "(unknown)");
}

static void
disconnect (NmnNetworkItem *item)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (item);

    if (!priv->ac) {
        g_warning ("Can not disconnect, the item is not active?");
        return;
    }

    nm_device_disconnect (priv->device, disconnect_cb, item);
}

void
nmn_network_item_connect_request (NmnNetworkItem *self)
{
    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    NMN_NETWORK_ITEM_GET_CLASS (self)->connect (self);
    g_signal_emit (self, signals[CONNECT_REQUESTED], 0);
}

void
nmn_network_item_disconnect_request (NmnNetworkItem *self)
{
    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    NMN_NETWORK_ITEM_GET_CLASS (self)->disconnect (self);
    g_signal_emit (self, signals[DISCONNECT_REQUESTED], 0);
}

void
nmn_network_item_set_security (NmnNetworkItem *self,
                               const char *security_string)
{
    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    gtk_label_set_text (GTK_LABEL (GET_PRIVATE (self)->security),
                        security_string ? security_string : "");
}

static void
nmn_network_item_delete (NmnNetworkItem *self)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (self);
    GtkDialog *dialog;
    GtkWidget *label;
    const char *name;
    char *label_text;

    dialog = GTK_DIALOG (gtk_dialog_new_with_buttons (_("Really remove?"),
                                                      NULL,
                                                      GTK_DIALOG_MODAL |
                                                      GTK_DIALOG_DESTROY_WITH_PARENT,
                                                      GTK_STOCK_CANCEL,
                                                      GTK_RESPONSE_REJECT,
                                                      GTK_STOCK_OK,
                                                      GTK_RESPONSE_ACCEPT,
                                                      NULL));

    gtk_dialog_set_has_separator (dialog, FALSE);
    gtk_dialog_set_default_response (dialog, GTK_RESPONSE_ACCEPT);
    gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
    gtk_window_set_icon_name (GTK_WINDOW (dialog), GTK_STOCK_DELETE);

    name = gtk_label_get_text (GTK_LABEL (priv->name));
    label_text = g_strdup_printf (_("Do you want to remove the '%s' %s network? "
                                    "This\nwill forget the password and you will"
                                    " no longer be\nautomatically connected to "
                                    "'%s'."),
                                  name, priv->type_str, name);
    label = gtk_label_new (label_text);

    gtk_box_set_spacing (GTK_BOX (dialog->vbox), 12);
    gtk_box_pack_start (GTK_BOX (dialog->vbox), label, TRUE, TRUE, 6);
    gtk_widget_show_all (GTK_WIDGET (dialog));

    if (gtk_dialog_run (dialog) == GTK_RESPONSE_ACCEPT)
        g_signal_emit (self, signals[DELETE], 0);

    gtk_widget_destroy (GTK_WIDGET (dialog));
}

void
nmn_network_item_set_delete_visible (NmnNetworkItem *item,
                                     gboolean visible)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (item);

    g_return_if_fail (NMN_IS_NETWORK_ITEM (item));

    if (visible && !priv->remove) {
        GtkWidget *button;

        priv->remove = gtk_vbox_new (FALSE, 0);
        gtk_box_pack_end (priv->hbox, priv->remove, FALSE, FALSE, 0);

        button = gtk_button_new ();
        gtk_button_set_image (GTK_BUTTON (button),
                              gtk_image_new_from_icon_name ("edit-clear", GTK_ICON_SIZE_MENU));

        gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);

        gtk_button_set_image_position (GTK_BUTTON (button), GTK_POS_RIGHT);
        gtk_widget_set_tooltip_text (button, "Remove connection");
        gtk_box_pack_start (GTK_BOX (priv->remove), button, TRUE, FALSE, 0);

        gtk_widget_show_all (priv->remove);
        g_signal_connect_swapped (button, "clicked",
                                  G_CALLBACK (nmn_network_item_delete),
                                  item);
    } else if (!visible && priv->remove) {
        gtk_container_remove (GTK_CONTAINER (priv->hbox), priv->remove);
        priv->remove = NULL;
    }
}

NMSettingsConnectionInterface *
nmn_network_item_get_connection (NmnNetworkItem *self)
{
    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (self), NULL);

    return GET_PRIVATE (self)->connection;
}

void
nmn_network_item_set_connection (NmnNetworkItem *self,
                                 NMSettingsConnectionInterface *connection)
{
    NmnNetworkItemPrivate *priv;

    g_return_if_fail (NMN_IS_NETWORK_ITEM (self));

    priv = GET_PRIVATE (self);

    if (priv->connection) {
        g_signal_handler_disconnect (priv->connection, priv->connection_updated_id);
        g_signal_handler_disconnect (priv->connection, priv->connection_removed_id);

        if (priv->secrets_requested_id)
            g_signal_handler_disconnect (priv->connection, priv->secrets_requested_id);

        g_object_unref (priv->connection);
    }

    if (connection) {
        priv->connection = g_object_ref (connection);
        priv->connection_updated_id = g_signal_connect (connection,
                                                        NM_SETTINGS_CONNECTION_INTERFACE_UPDATED,
                                                        G_CALLBACK (updated),
                                                        self);
        priv->connection_removed_id = g_signal_connect (connection,
                                                        NM_SETTINGS_CONNECTION_INTERFACE_REMOVED,
                                                        G_CALLBACK (removed),
                                                        self);

        if (NMA_IS_GCONF_CONNECTION (connection))
            priv->secrets_requested_id = g_signal_connect (connection, "new-secrets-requested",
                                                           G_CALLBACK (connection_secrets_requested_cb), self);
        else
            priv->secrets_requested_id = 0;

        update_item (self);
    } else {
        priv->connection = NULL;
        priv->connection_updated_id = 0;
    }
}

static void
delete_cb (NMSettingsConnectionInterface *connection,
           GError *error,
           gpointer user_data)
{
    if (error)
        g_warning ("Error deleting connection: (%d) %s", error->code, error->message);
    if (user_data)
        g_object_set_data (G_OBJECT (user_data), "conn", NULL);
}

static void
item_delete (NmnNetworkItem *item)
{
    NMSettingsConnectionInterface *connection = nmn_network_item_get_connection (item);

    if (connection)
    	nm_settings_connection_interface_delete (connection, delete_cb, NULL);
}

static guint
get_priority (NmnNetworkItem *item)
{
    NMActiveConnection *ac;
    guint priority = 0;

    ac = nmn_network_item_get_active_connection (item);
    if (ac) {
        priority = 90;

        if (nm_active_connection_get_default (ac))
            priority += 10;
    }

    return priority;
}


static void
details_changed (NmnConnectionDetails *details,
                 gboolean complete,
                 gpointer user_data)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (user_data);

    gtk_widget_set_sensitive (GTK_WIDGET (priv->connect), complete);
}

gboolean
nmn_network_item_has_details (NmnNetworkItem *item)
{
    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (item), FALSE);

    return GET_PRIVATE (item)->details != NULL;
}

NmnConnectionDetails *
nmn_network_item_get_details (NmnNetworkItem *item)
{
    NmnNetworkItemPrivate *priv;
    GtkWidget *alignment;

    g_return_val_if_fail (NMN_IS_NETWORK_ITEM (item), NULL);

    priv = GET_PRIVATE (item);
    if (priv->details)
        return priv->details;

    priv->details = nmn_connection_details_new ();
    update_details (item);

    alignment = gtk_alignment_new (0, 0, 0, 0);
    gtk_container_add (GTK_CONTAINER (alignment), GTK_WIDGET (priv->details));
    gtk_widget_show (alignment);
    gtk_box_pack_end (priv->vbox, alignment, FALSE, FALSE, 0);

    g_signal_connect (priv->details, "changed", G_CALLBACK (details_changed), item);
    details_changed (priv->details, nmn_connection_details_verify (priv->details), item);

    return priv->details;
}

static void
advanced_expanded (GtkExpander *expander,
                   GParamSpec *param_spec,
                   gpointer user_data)
{
    NmnConnectionDetails *details;

    details = nmn_network_item_get_details (NMN_NETWORK_ITEM (user_data));
    if (details)
        g_object_set (details, "visible", gtk_expander_get_expanded (expander), NULL);
}

static void
connect_clicked (GtkButton *button, gpointer user_data)
{
    NmnNetworkItem *item = NMN_NETWORK_ITEM (user_data);
    NmnNetworkItemPrivate *priv = GET_PRIVATE (item);

    if (priv->status == NMN_NETWORK_ITEM_STATUS_DISCONNECTED)
        nmn_network_item_connect_request (item);
    else
        nmn_network_item_disconnect_request (item);
}

static guint
nmn_network_item_get_priority (NmnNetworkItem *self)
{
    return NMN_NETWORK_ITEM_GET_CLASS (self)->get_priority (self);
}

static gboolean
enter_notify_event (GtkWidget *widget,
                    GdkEventCrossing *event)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (widget);

    gtk_widget_modify_bg (widget, GTK_STATE_NORMAL, &priv->prelight_color);

    return TRUE;
}

static gboolean
leave_notify_event (GtkWidget *widget,
                    GdkEventCrossing *event)
{
    if (event->detail != GDK_NOTIFY_INFERIOR)
        gtk_widget_modify_bg (widget, GTK_STATE_NORMAL, NULL);

    return TRUE;
}

static void
item_init (NmnItem *item_class)
{
}

static void
nmn_network_item_init (NmnNetworkItem *item)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (item);
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *w;

    priv->vbox = GTK_BOX (gtk_vbox_new (FALSE, 6));
    gtk_container_set_border_width (GTK_CONTAINER (priv->vbox), 6);
    gtk_container_add (GTK_CONTAINER (item), GTK_WIDGET (priv->vbox));

    priv->hbox = GTK_BOX (gtk_hbox_new (FALSE, 6));
    gtk_container_set_border_width (GTK_CONTAINER (priv->hbox), 6);
    gtk_box_pack_start (priv->vbox, GTK_WIDGET (priv->hbox), FALSE, FALSE, 0);

    priv->icon = gtk_image_new ();
    gtk_box_pack_start (priv->hbox, priv->icon, FALSE, FALSE, 0);

    vbox = gtk_vbox_new (FALSE, 0);
    gtk_box_pack_start (priv->hbox, vbox, FALSE, FALSE, 0);

    hbox = gtk_hbox_new (FALSE, 12);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

    w = gtk_label_new ("");
    gtk_label_set_use_markup (GTK_LABEL (w), TRUE);
    gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
    gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 0);
    priv->name = GTK_LABEL (w);

    w = gtk_label_new (" - ");
    gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 0);
    priv->status_separator = w;

    w = gtk_label_new ("");
    gtk_label_set_use_markup (GTK_LABEL (w), TRUE);
    gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
    gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 0);
    priv->status_label = GTK_LABEL (w);

    hbox = gtk_hbox_new (FALSE, 12);
    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

    w = gtk_label_new ("");
    gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 0);
    priv->security = GTK_LABEL (w);

    w = gtk_expander_new (_("Advanced"));
    gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 0);
    priv->expander = w;
    g_signal_connect (w, "notify::expanded", G_CALLBACK (advanced_expanded), item);

    w = gtk_button_new_with_label (_("Connect"));
    gtk_box_pack_start (GTK_BOX (hbox), w, FALSE, FALSE, 0);
    priv->connect = GTK_BUTTON (w);
    g_signal_connect (w, "clicked", G_CALLBACK (connect_clicked), item);

    priv->remove = NULL;

    gdk_color_parse ("#cbcbcb", &priv->prelight_color);

    gtk_widget_show_all (GTK_WIDGET (priv->vbox));
}

static GObject*
constructor (GType type,
             guint n_construct_params,
             GObjectConstructParam *construct_params)
{
    GObject *object;
    NmnNetworkItemPrivate *priv;

    object = G_OBJECT_CLASS (nmn_network_item_parent_class)->constructor
        (type, n_construct_params, construct_params);

    if (!object)
        return NULL;

    priv = GET_PRIVATE (object);

    if (!priv->nm_data || !priv->device || !priv->type_str) {
        g_warning ("%s: Missing constructor arguments", G_STRLOC);
        g_object_unref (object);
        return NULL;
    }

    return object;
}

static void
set_property (GObject *object, guint prop_id,
              const GValue *value, GParamSpec *pspec)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (object);
    char *s;

    switch (prop_id) {
    case NMN_ITEM_PROP_NAME:
        s = g_strdup_printf ("<big><b>%s</b></big>", g_value_get_string (value));
        gtk_label_set_markup (GTK_LABEL (priv->name), s);
        g_free (s);
        g_object_notify (object, NMN_ITEM_NAME);
        break;
    case PROP_NM_DATA:
        /* Construct only */
        priv->nm_data = g_value_dup_object (value);
        break;
    case PROP_DEVICE:
        /* Construct only */
        priv->device = g_value_dup_object (value);
        break;
    case PROP_CONNECTION:
        nmn_network_item_set_connection (NMN_NETWORK_ITEM (object),
                                         (NMSettingsConnectionInterface *) g_value_get_object (value));
        break;
    case PROP_AC:
        nmn_network_item_set_active_connection (NMN_NETWORK_ITEM (object),
                                                (NMActiveConnection *) g_value_get_object (value));
        break;
    case PROP_TYPE_STRING:
        /* Construct only */
        priv->type_str = g_value_dup_string (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
get_property (GObject *object, guint prop_id,
              GValue *value, GParamSpec *pspec)
{
    NmnNetworkItemPrivate *priv = GET_PRIVATE (object);

    switch (prop_id) {
    case NMN_ITEM_PROP_NAME:
        g_value_set_string (value, gtk_label_get_text (GTK_LABEL (priv->name)));
        break;
    case NMN_ITEM_PROP_PRIORITY:
        g_value_set_uint (value, nmn_network_item_get_priority (NMN_NETWORK_ITEM (object)));
        break;
    case PROP_NM_DATA:
        g_value_set_object (value, priv->nm_data);
        break;
    case PROP_DEVICE:
        g_value_set_object (value, priv->device);
        break;
    case PROP_CONNECTION:
        g_value_set_object (value, priv->connection);
        break;
    case PROP_AC:
        g_value_set_object (value, priv->ac);
        break;
    case PROP_TYPE_STRING:
        g_value_set_string (value, priv->type_str);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
dispose (GObject *object)
{
    NmnNetworkItem *self = NMN_NETWORK_ITEM (object);
    NmnNetworkItemPrivate *priv = GET_PRIVATE (object);

    if (priv->disposed)
        return;

    g_free (priv->type_str);

    nmn_network_item_set_active_connection (self, NULL);

    if (priv->device)
        g_object_unref (priv->device);

    nmn_network_item_set_connection (self, NULL);

    if (priv->nm_data)
        g_object_unref (priv->nm_data);

    priv->disposed = TRUE;

    G_OBJECT_CLASS (nmn_network_item_parent_class)->dispose (object);
}

static void
nmn_network_item_class_init (NmnNetworkItemClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

    g_type_class_add_private (object_class, sizeof (NmnNetworkItemPrivate));

    object_class->constructor = constructor;
    object_class->set_property = set_property;
    object_class->get_property = get_property;
    object_class->dispose = dispose;

    class->connect = connect;
    class->disconnect = disconnect;
    class->delete = item_delete;
    class->get_priority = get_priority;

    widget_class->enter_notify_event = enter_notify_event;
    widget_class->leave_notify_event = leave_notify_event;

    /* properties */
    g_object_class_override_property (object_class,
                                      NMN_ITEM_PROP_NAME,
                                      NMN_ITEM_NAME);

    g_object_class_override_property (object_class,
                                      NMN_ITEM_PROP_PRIORITY,
                                      NMN_ITEM_PRIORITY);

    g_object_class_install_property
        (object_class, PROP_NM_DATA,
         g_param_spec_object (NMN_NETWORK_ITEM_NM_DATA,
                              "NmnNMData",
                              "NmnNMData",
                              NMN_TYPE_NM_DATA,
                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    g_object_class_install_property
        (object_class, PROP_DEVICE,
         g_param_spec_object (NMN_NETWORK_ITEM_DEVICE,
                              "NMDevice",
                              "NMDevice",
                              NM_TYPE_DEVICE,
                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    g_object_class_install_property
        (object_class, PROP_CONNECTION,
         g_param_spec_object (NMN_NETWORK_ITEM_CONNECTION,
                              "NMSettingsConnectionInterface",
                              "NMSettingsConnectionInterface",
                              NM_TYPE_SETTINGS_CONNECTION_INTERFACE,
                              G_PARAM_READWRITE));

    g_object_class_install_property
        (object_class, PROP_AC,
         g_param_spec_object (NMN_NETWORK_ITEM_AC,
                              "NMActiveConnection",
                              "NMActiveConnection",
                              NM_TYPE_ACTIVE_CONNECTION,
                              G_PARAM_READWRITE));

    g_object_class_install_property
        (object_class, PROP_TYPE_STRING,
         g_param_spec_string (NMN_NETWORK_ITEM_TYPE_STRING,
                              "Type string",
                              "Type string",
                              NULL,
                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

    /* Signals */
    signals[CONNECT_REQUESTED] = g_signal_new 
        ("connect-requested",
         G_OBJECT_CLASS_TYPE (class),
         G_SIGNAL_RUN_LAST,
         G_STRUCT_OFFSET (NmnNetworkItemClass, connect_requested),
         NULL, NULL,
         g_cclosure_marshal_VOID__VOID,
         G_TYPE_NONE, 0);

    signals[DISCONNECT_REQUESTED] = g_signal_new 
        ("disconnect-requested",
         G_OBJECT_CLASS_TYPE (class),
         G_SIGNAL_RUN_LAST,
         G_STRUCT_OFFSET (NmnNetworkItemClass, disconnect_requested),
         NULL, NULL,
         g_cclosure_marshal_VOID__VOID,
         G_TYPE_NONE, 0);

    signals[DELETE] = g_signal_new 
        ("delete",
         G_OBJECT_CLASS_TYPE (class),
         G_SIGNAL_RUN_LAST,
         G_STRUCT_OFFSET (NmnNetworkItemClass, delete),
         NULL, NULL,
         g_cclosure_marshal_VOID__VOID,
         G_TYPE_NONE, 0);
}
