/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2009 Novell, Inc.
 */

#include <glib/gi18n.h>
#include <NetworkManager.h>
#include <nm-connection.h>
#include <nm-setting-connection.h>
#include <nm-setting-wired.h>
#include "nmn-serial-item.h"
#include "nma-gconf-connection.h"
#include "utils.h"

G_DEFINE_TYPE (NmnSerialItem, nmn_serial_item, NMN_TYPE_NETWORK_ITEM)

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NMN_TYPE_SERIAL_ITEM, NmnSerialItemPrivate))

typedef struct {
    gulong state_changed_id;

    gboolean disposed;
} NmnSerialItemPrivate;

static void
state_changed (NMDevice *device,
               NMDeviceState new_state,
               NMDeviceState old_state,
               NMDeviceStateReason reason,
               gpointer user_data)
{
    const char *icon;

    switch (new_state) {
    case NM_DEVICE_STATE_ACTIVATED:
        icon = "nm-device-wwan";
        break;
    default:
        icon = "nm-no-connection";
        break;
    }

    nmn_network_item_set_icon (NMN_NETWORK_ITEM (user_data), icon);
}

GtkWidget *
nmn_serial_item_new (NmnNMData *nm_data,
                     NMSerialDevice *device)
{
    g_return_val_if_fail (NMN_IS_NM_DATA (nm_data), NULL);
    g_return_val_if_fail (NM_IS_SERIAL_DEVICE (device), NULL);

    return GTK_WIDGET (g_object_new (NMN_TYPE_SERIAL_ITEM,
                                     NMN_NETWORK_ITEM_NM_DATA, nm_data,
                                     NMN_NETWORK_ITEM_DEVICE, device,
                                     NMN_NETWORK_ITEM_TYPE_STRING, _("3G"),
                                     NULL));
}

static guint
get_priority (NmnNetworkItem *item)
{
    return NMN_NETWORK_ITEM_CLASS (nmn_serial_item_parent_class)->get_priority (item) + 10;
}

static gboolean
set_name (gpointer data)
{
    NMSettingsConnectionInterface *connection;
    NMSettingConnection *s_con;

    connection = nmn_network_item_get_connection (NMN_NETWORK_ITEM (data));
    s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (NM_CONNECTION (connection), NM_TYPE_SETTING_CONNECTION));

    if (s_con)
        nmn_item_set_name (NMN_ITEM (data), nm_setting_connection_get_id (s_con));

    return FALSE;
}

static void
nmn_serial_item_init (NmnSerialItem *item)
{
    nmn_network_item_set_delete_visible (NMN_NETWORK_ITEM (item), TRUE);
}

static GObject*
constructor (GType type,
             guint n_construct_params,
             GObjectConstructParam *construct_params)
{
    GObject *object;
    NmnSerialItemPrivate *priv;
    NMDevice *device;

    object = G_OBJECT_CLASS (nmn_serial_item_parent_class)->constructor
        (type, n_construct_params, construct_params);

    if (!object)
        return NULL;

    priv = GET_PRIVATE (object);
    device = nmn_network_item_get_device (NMN_NETWORK_ITEM (object));
    priv->state_changed_id = g_signal_connect (device, "state-changed", G_CALLBACK (state_changed), object);

    state_changed (device, nm_device_get_state (device), 
                   NM_DEVICE_STATE_UNKNOWN, NM_DEVICE_STATE_REASON_NONE, object);

    g_idle_add (set_name, object);

    return object;
}

static void
dispose (GObject *object)
{
    NmnSerialItemPrivate *priv = GET_PRIVATE (object);
    NMDevice *device;

    if (priv->disposed)
        return;

    device = nmn_network_item_get_device (NMN_NETWORK_ITEM (object));
    g_signal_handler_disconnect (device, priv->state_changed_id);

    priv->disposed = TRUE;

    G_OBJECT_CLASS (nmn_serial_item_parent_class)->dispose (object);
}

static void
nmn_serial_item_class_init (NmnSerialItemClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    NmnNetworkItemClass *network_item_class = NMN_NETWORK_ITEM_CLASS (class);

    g_type_class_add_private (object_class, sizeof (NmnSerialItemPrivate));

    object_class->constructor = constructor;
    object_class->dispose = dispose;

    network_item_class->get_priority = get_priority;
}
