/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2009 Novell, Inc.
 */

#include <string.h>
#include <glib/gi18n.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <libnotify/notify.h>
#include <nm-device-ethernet.h>
#include <nm-device-wifi.h>
#include <nm-gsm-device.h>
#include <nm-cdma-device.h>
#include <nm-utils.h>

#include "nmn-status-icon.h"
#include "nmn-icon-cache.h"

#define ACTIVATION_STEPS 6

G_DEFINE_TYPE (NmnStatusIcon, nmn_status_icon, G_TYPE_OBJECT)

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NMN_TYPE_STATUS_ICON, NmnStatusIconPrivate))

enum {
    STATUS_DISCONNECTED,
    STATUS_CONNECTING,
    STATUS_CONNECTED
};

typedef struct {
    NmnStatusIcon *icon;
    void (*destroy) (gpointer);
} NmnIconHandler;

typedef struct {
    NMClient *nm_client;
    MplPanelClient *panel_client;

    GSList *ac_list;
    NmnIconHandler *ac_icon_handler;

    /* Notification ... and state */
    char *connection_type;
    char *connection_name;
    int status;
} NmnStatusIconPrivate;

NmnStatusIcon *
nmn_status_icon_new (void)
{
    return NMN_STATUS_ICON (g_object_new (NMN_TYPE_STATUS_ICON, NULL));
}

static void
update_icon (NmnStatusIcon *self, const char *image)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);

    if (priv->panel_client)
        mpl_panel_client_request_button_style (priv->panel_client, image);
}

static void
update_notification (NmnStatusIcon *self,
                     int status,
                     const char *connection_type,
                     const char *connection_name,
                     const char *icon)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    const char *title;
    char *msg;

    if (priv->status == STATUS_DISCONNECTED && status == STATUS_CONNECTED) {
        title = _("Network connected");

        if (connection_type) {
            if (connection_name)
                msg = g_strdup_printf (_("You're now connected to %s, a %s network"),
                                       connection_name, connection_type);
            else
                msg = g_strdup_printf (_("You're now connected to %s network"), connection_type);
        } else
            msg = g_strdup (_("You're now connected to network"));
    } else if (priv->status == STATUS_CONNECTED && status == STATUS_DISCONNECTED) {
        title = _("Network lost");

        if (priv->connection_type) {
            if (priv->connection_name)
                msg = g_strdup_printf (_("Sorry, we've lost your %s connection to %s"),
                                       priv->connection_type, priv->connection_name);
            else
                msg = g_strdup_printf (_("Sorry, we've lost your %s connection"), priv->connection_type);
        } else
            msg = g_strdup (_("Sorry, we've lost your connection"));
    } else {
        title = NULL;
        msg = NULL;
    }

    if (title && msg) {
        NotifyNotification *note;

        note = notify_notification_new (title, msg, icon, NULL);
        notify_notification_set_timeout (note, 10000);
        notify_notification_show (note, NULL);
        g_object_unref (note);
        g_free (msg);

        priv->status = status;
        g_free (priv->connection_type);
        priv->connection_type = connection_type ? g_strdup (connection_type) : NULL;
        g_free (priv->connection_name);
        priv->connection_name = connection_name ? g_strdup (connection_name) : NULL;
    }
}

static void
update_tooltip (NmnStatusIcon *self,
                int status,
                const char *connection_type,
                const char *connection_name)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    char *tip;

    switch (status) {
    case STATUS_DISCONNECTED:
        tip = g_strdup_printf (_("networks - not connected"));
        break;
    case STATUS_CONNECTING:
        tip = g_strdup_printf (_("networks - connecting"));
        break;
    case STATUS_CONNECTED:
        if (connection_type) {
            if (connection_name)
                tip = g_strdup_printf (_("networks - %s - %s"), connection_name, connection_type);
            else
                tip = g_strdup_printf (_("networks - %s"), connection_type);
        } else
            tip = g_strdup (_("networks - connected"));

        break;
    }

    mpl_panel_client_request_tooltip (priv->panel_client, tip);
    g_free (tip);
}

static void
update (NmnStatusIcon *self,
        int status,
        const char *connection_type,
        const char *connection_name,
        const char *icon)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);

    update_notification (self, status, connection_type, connection_name, icon);

    if (priv->panel_client) {
        update_icon (self, icon);
        update_tooltip (self, status, connection_type, connection_name);
    }
}

static void
set_status_disconnected (NmnStatusIcon *self)
{
    update (self, STATUS_DISCONNECTED, NULL, NULL, "no-network");
}

static void
set_status_connecting (NmnStatusIcon *self)
{
    update_tooltip (self, STATUS_CONNECTING, NULL, NULL);
}

static void
set_status_connected (NmnStatusIcon *self,
                      const char *connection_type,
                      const char *connection_name,
                      const char *icon)
{
    update (self, STATUS_CONNECTED, connection_type, connection_name, icon);
}

static void
icon_handler_destroy (NmnIconHandler *handler)
{
    g_return_if_fail (handler != NULL);

    if (handler->destroy)
        handler->destroy (handler);
    else
        g_warning ("Missing destroy method");
}

/* Pending activation animation handler */

typedef struct {
    NmnIconHandler parent;
    guint step;
    guint animation_id;
    guint animation_index;
} NmnPendingActivationHandler;

static void
icon_handler_pending_activation_destroy (gpointer data)
{
    NmnPendingActivationHandler *handler = (NmnPendingActivationHandler *) data;

    g_source_remove (handler->animation_id);
    g_free (handler);
}

static gboolean
activation_animation (gpointer data)
{
    NmnPendingActivationHandler *handler = (NmnPendingActivationHandler *) data;
    char *image;

    if (++handler->step > ACTIVATION_STEPS)
        handler->step = 1;

    image = g_strdup_printf ("progress-%02d", handler->step);
    update_icon (handler->parent.icon, image);
    g_free (image);

    return TRUE;
}

static NmnIconHandler *
icon_handler_pending_activation_new (NmnStatusIcon *icon)
{
    NmnPendingActivationHandler *handler;

    handler = g_new0 (NmnPendingActivationHandler, 1);
    handler->parent.icon = icon;
    handler->parent.destroy = icon_handler_pending_activation_destroy;

    handler->animation_id = g_timeout_add (200, activation_animation, handler);
    activation_animation (handler);
    set_status_connecting (icon);

    return (NmnIconHandler *) handler;
}


/* Ethernet handler */

static NmnIconHandler *
icon_handler_ethernet_new (NmnStatusIcon *icon)
{
    NmnIconHandler *handler;

    handler = g_new0 (NmnIconHandler, 1);
    handler->destroy = g_free;
    set_status_connected (icon, _("wired"), NULL, "device-wired");

    return handler;
}


/* Wifi handler */

typedef struct {
    NmnIconHandler parent;

    NMDevice *device;
    gulong active_ap_id;

    NMAccessPoint *ap;
    gulong ap_strength_id;

    gboolean just_activated;
} NmnWifiHandler;

static void
icon_handler_wifi_destroy (gpointer data)
{
    NmnWifiHandler *handler = (NmnWifiHandler *) data;

    g_signal_handler_disconnect (handler->device, handler->active_ap_id);

    if (handler->ap_strength_id)
        g_signal_handler_disconnect (handler->ap, handler->ap_strength_id);

    g_free (handler);
}

static void
ap_strength_changed (NMAccessPoint *ap, GParamSpec *pspec, gpointer user_data)
{
    NmnIconHandler *handler = (NmnIconHandler *) user_data;
    const char *image;
    guint32 strength;

    /* Hack: libnm-glib emits this signal sometimes with AP == NULL.
       Should fix that instead */
    strength = ap ? nm_access_point_get_strength (ap) : 0;
    strength = CLAMP (strength, 0, 100);

    if (strength > 80)
        image = "signal-100";
    else if (strength > 55)
        image = "signal-75";
    else if (strength > 30)
        image = "signal-50";
    else if (strength > 5)
        image = "signal-25";
    else
        image = "signal-00";

    update_icon (handler->icon, image);
}

static void
active_ap_changed (NMDeviceWifi *device, GParamSpec *pspec, gpointer user_data)
{
    NmnWifiHandler *handler = (NmnWifiHandler *) user_data;
    NmnStatusIcon *icon = ((NmnIconHandler *) handler)->icon;

    if (handler->ap_strength_id) {
        g_signal_handler_disconnect (handler->ap, handler->ap_strength_id);

        handler->ap = NULL;
        handler->ap_strength_id = 0;
    }

    handler->ap = nm_device_wifi_get_active_access_point (device);
    if (handler->ap) {
        const GByteArray *ssid;
        char *id;
        const char *connection_type = _("WiFi");

        ssid = nm_access_point_get_ssid (handler->ap);
        if (ssid)
            id = nm_utils_ssid_to_utf8 ((char *) ssid->data, ssid->len);
        else
            id = g_strdup (_("(Hidden"));

        if (handler->just_activated) {
            update_notification (icon, STATUS_CONNECTED, connection_type, id, "nm-device-wifi");
            handler->just_activated = FALSE;
        }

        update_tooltip (icon, STATUS_CONNECTED, connection_type, id);
        g_free (id);

        handler->ap_strength_id = g_signal_connect (handler->ap, "notify::" NM_ACCESS_POINT_STRENGTH,
                                                    G_CALLBACK (ap_strength_changed), handler);
    }

    ap_strength_changed (handler->ap, NULL, handler);
}

static NmnIconHandler *
icon_handler_wifi_new (NmnStatusIcon *icon, NMActiveConnection *ac)
{
    const GPtrArray *active_devices;
    NMDevice *device;
    NmnWifiHandler *handler;

    active_devices = nm_active_connection_get_devices (ac);
    if (!active_devices)
        return NULL;

    device = NM_DEVICE (g_ptr_array_index (active_devices, 0));

    handler = g_new0 (NmnWifiHandler, 1);
    handler->parent.icon = icon;
    handler->parent.destroy = icon_handler_wifi_destroy;
    handler->just_activated = TRUE;

    handler->device = device;
    handler->active_ap_id = g_signal_connect (device, "notify::" NM_DEVICE_WIFI_ACTIVE_ACCESS_POINT,
                                              G_CALLBACK (active_ap_changed),
                                              handler);

    active_ap_changed (NM_DEVICE_WIFI (device), NULL, handler);

    return (NmnIconHandler *) handler;
}


/* 3G handler */

static NmnIconHandler *
icon_handler_3g_new (NmnStatusIcon *icon)
{
    NmnIconHandler *handler;

    handler = g_new0 (NmnIconHandler, 1);
    handler->destroy = g_free;

    /* FIXME: Get the connection name */
    set_status_connected (icon, _("3G"), NULL, "device-wwan");

    return handler;
}


/* Handler creator */

static NmnIconHandler *
icon_handler_create (NmnStatusIcon *self, NMActiveConnection *ac)
{
    NmnIconHandler *handler = NULL;
    NMActiveConnectionState state;

    state = nm_active_connection_get_state (ac);

    if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
        handler = icon_handler_pending_activation_new (self);
    else if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATED) {
        const GPtrArray *active_devices;
        NMDevice *device = NULL;

        active_devices = nm_active_connection_get_devices (ac);
        if (active_devices)
            device = NM_DEVICE (g_ptr_array_index (active_devices, 0));

        if (NM_IS_DEVICE_ETHERNET (device))
            handler = icon_handler_ethernet_new (self);
        else if (NM_IS_DEVICE_WIFI (device))
            handler = icon_handler_wifi_new (self, ac);
        else if (NM_IS_GSM_DEVICE (device) || NM_IS_CDMA_DEVICE (device))
            handler = icon_handler_3g_new (self);
        else
            g_warning ("Unhandled device type: '%s'", G_OBJECT_TYPE_NAME (device));
    }

    return handler;
}

typedef struct {
    NMActiveConnection *ac;
    gulong signal_id;
} AcListItem;

static void
update_best_ac (NmnStatusIcon *self)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    GSList *iter;
    NMActiveConnection *best_ac = NULL;

    for (iter = priv->ac_list; iter; iter = iter->next) {
        AcListItem *item = (AcListItem *) iter->data;

        if (nm_active_connection_get_state (item->ac) == NM_ACTIVE_CONNECTION_STATE_UNKNOWN)
            continue;

        if (!best_ac ||
            nm_active_connection_get_state (item->ac) > nm_active_connection_get_state (best_ac) ||
            nm_active_connection_get_default (item->ac) == TRUE)
            best_ac = item->ac;
    }

    if (priv->ac_icon_handler) {
        icon_handler_destroy (priv->ac_icon_handler);
        priv->ac_icon_handler = NULL;
    }

    if (best_ac)
        priv->ac_icon_handler = icon_handler_create (self, best_ac);

    if (!priv->ac_icon_handler)
        set_status_disconnected (self);
}

static void
ac_state_changed (NMActiveConnection *ac,
                  GParamSpec *pspec,
                  gpointer user_data)
{
    update_best_ac (NMN_STATUS_ICON (user_data));
}

static void
ac_list_destroy (NmnStatusIcon *self)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);

    while (priv->ac_list) {
        AcListItem *item = (AcListItem *) priv->ac_list->data;

        g_signal_handler_disconnect (item->ac, item->signal_id);
        g_object_unref (item->ac);
        g_free (item);
        priv->ac_list = g_slist_delete_link (priv->ac_list, priv->ac_list);
    }
}

static void
active_connections_changed (NMClient *client,
                            GParamSpec *pspec,
                            gpointer user_data)
{
    NmnStatusIcon *self = NMN_STATUS_ICON (user_data);
    NmnStatusIconPrivate *priv = GET_PRIVATE (self);
    const GPtrArray *ac_list;
    int i;

    /* First, disconnect all previous handlers */
    ac_list_destroy (self);

    /* add signal handlers for each AC */
    if (nm_client_get_manager_running (client))
        ac_list = nm_client_get_active_connections (client);
    else
        ac_list = NULL;

    for (i = 0; ac_list && i < ac_list->len; i++) {
        NMActiveConnection *ac = (NMActiveConnection *) g_ptr_array_index (ac_list, i);
        AcListItem *item;

        item = g_new (AcListItem, 1);
        item->ac = g_object_ref (ac);
        item->signal_id = g_signal_connect (ac, "notify", G_CALLBACK (ac_state_changed), self);
        priv->ac_list = g_slist_prepend (priv->ac_list, item);
    }

    update_best_ac (self);
}

void
nmn_status_icon_set_nm_client (NmnStatusIcon *self,
                               NMClient *client)
{
    NmnStatusIconPrivate *priv;

    g_return_if_fail (NMN_IS_STATUS_ICON (self));
    g_return_if_fail (NM_IS_CLIENT (client));

    priv = GET_PRIVATE (self);
    priv->nm_client = g_object_ref (client);

    g_signal_connect (priv->nm_client, "notify::" NM_CLIENT_ACTIVE_CONNECTIONS,
	                  G_CALLBACK (active_connections_changed),
	                  self);

    active_connections_changed (client, NULL, self);
}

void
nmn_status_icon_set_panel_client (NmnStatusIcon *self,
                                  MplPanelClient *panel_client)
{
    NmnStatusIconPrivate *priv;

    g_return_if_fail (NMN_IS_STATUS_ICON (self));
    g_return_if_fail (MPL_IS_PANEL_CLIENT (panel_client));

    priv = GET_PRIVATE (self);
    if (priv->panel_client)
        g_object_unref (priv->panel_client);

    priv->panel_client = g_object_ref (panel_client);

    set_status_disconnected (self);
}

/*****************************************************************************/

static gboolean
set_initial_icon (gpointer data)
{
    NmnStatusIcon *self = NMN_STATUS_ICON (data);

    notify_init ("network-manager-netbook");
    set_status_disconnected (self);

    return FALSE;
}

static void
nmn_status_icon_init (NmnStatusIcon *self)
{
    g_idle_add (set_initial_icon, self);
}

static void
finalize (GObject *object)
{
    NmnStatusIconPrivate *priv = GET_PRIVATE (object);

    g_free (priv->connection_type);
    g_free (priv->connection_name);

    if (priv->ac_icon_handler)
        icon_handler_destroy (priv->ac_icon_handler);

    ac_list_destroy (NMN_STATUS_ICON (object));

    if (priv->panel_client)
        g_object_unref (priv->panel_client);

    if (priv->nm_client)
        g_object_unref (priv->nm_client);

    notify_uninit ();

    G_OBJECT_CLASS (nmn_status_icon_parent_class)->finalize (object);
}

static void
nmn_status_icon_class_init (NmnStatusIconClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);

    g_type_class_add_private (object_class, sizeof (NmnStatusIconPrivate));

    object_class->finalize = finalize;
}
