/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2009 Novell, Inc.
 */

#include <string.h>
#include <nm-setting-connection.h>
#include <nm-setting-wireless.h>
#include "nmn-wifi-handler.h"
#include "nmn-wifi-item.h"
#include "utils.h"

G_DEFINE_TYPE (NmnWifiHandler, nmn_wifi_handler, NMN_TYPE_DEVICE_HANDLER)

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NMN_TYPE_WIFI_HANDLER, NmnWifiHandlerPrivate))

typedef struct {
    gulong wifi_toggled_id;
    gulong ap_added_id;
    gulong ap_removed_id;
    gulong ap_changed_id;

    gboolean disposed;
} NmnWifiHandlerPrivate;

NmnDeviceHandler *
nmn_wifi_handler_new (NmnNMData *nm_data,
                      NMDeviceWifi *device)
{
    g_return_val_if_fail (NMN_IS_NM_DATA (nm_data), NULL);
    g_return_val_if_fail (NM_IS_DEVICE_WIFI (device), NULL);

    return NMN_DEVICE_HANDLER (g_object_new (NMN_TYPE_WIFI_HANDLER,
                                             NMN_DEVICE_HANDLER_NM_DATA, nm_data,
                                             NMN_DEVICE_HANDLER_DEVICE, device,
                                             NULL));
}

static gboolean
aps_are_compatible (NMAccessPoint *ap1, NMAccessPoint *ap2)
{
    const GByteArray *ssid1;
    const GByteArray *ssid2;

    ssid1 = nm_access_point_get_ssid (ap1);
    ssid2 = nm_access_point_get_ssid (ap2);

    /* SSIDs do not match */
    if (!ssid1 || !ssid2 || ssid1->len != ssid2->len || memcmp (ssid1->data, ssid2->data, ssid1->len))
        return FALSE;

    if (nm_access_point_get_mode (ap1) == nm_access_point_get_mode (ap2) &&
        nm_access_point_get_flags (ap1) == nm_access_point_get_flags (ap2) &&
        nm_access_point_get_wpa_flags (ap1) == nm_access_point_get_wpa_flags (ap2) &&
        nm_access_point_get_rsn_flags (ap1) == nm_access_point_get_rsn_flags (ap2))
        return TRUE;

    return FALSE;
}

static NMAccessPoint *
find_best_compatible_ap (const GPtrArray *aps, NMAccessPoint *current_ap)
{
    NMAccessPoint *best_ap = NULL;
    int i;

    for (i = 0; aps && i < aps->len; i++) {
        NMAccessPoint *ap = NM_ACCESS_POINT (g_ptr_array_index (aps, i));

        if (ap == current_ap || !aps_are_compatible (current_ap, ap))
            continue;

        if (!best_ap || nm_access_point_get_strength (best_ap) < nm_access_point_get_strength (ap))
            best_ap = ap;
    }

    return best_ap;
}

static NMAccessPoint *
find_best_ap_for_connection (NMDeviceWifi *device,
                             NMConnection *connection)
{
    const GPtrArray *aps;
    NMAccessPoint *best_ap = NULL;
    int i;

    aps = nm_device_wifi_get_access_points (device);
    for (i = 0; aps && i < aps->len; i++) {
        NMAccessPoint *ap = NM_ACCESS_POINT (g_ptr_array_index (aps, i));

        if (utils_connection_valid_for_device (connection, NM_DEVICE (device), ap)) {
            if (!best_ap)
                best_ap = ap;
            else if (nm_access_point_get_strength (best_ap) < nm_access_point_get_strength (ap))
                best_ap = ap;
        }
    }

    return best_ap;
}

static void
ap_updated (NMAccessPoint *ap,
            GParamSpec *pspec,
            gpointer user_data)
{
    NmnDeviceHandler *handler = NMN_DEVICE_HANDLER (user_data);
    GSList *items;
    GSList *iter;

    items = nmn_device_handler_get_items (handler);
    for (iter = items; iter; iter = iter->next) {
        NmnWifiItem *item = NMN_WIFI_ITEM (iter->data);
        NMAccessPoint *item_ap = nmn_wifi_item_get_ap (item);
        NMDeviceWifi *device;
        NMAccessPoint *new_ap;

        /* Don't touch activated items, these are handled by active_ap_changed */
        if (nmn_network_item_get_status (NMN_NETWORK_ITEM (item)) != NMN_NETWORK_ITEM_STATUS_DISCONNECTED)
            continue;

        /* not us */
        if (item_ap != ap && !aps_are_compatible (ap, item_ap))
            continue;

        /* Do we have a better compatible AP for this item? */
        device = NM_DEVICE_WIFI (nmn_device_handler_get_device (handler));
        new_ap = find_best_compatible_ap (nm_device_wifi_get_access_points (device), item_ap);
        if (new_ap)
            nmn_wifi_item_set_ap (item, new_ap);
    }
}

static void
ap_removed (NMDeviceWifi *device,
            NMAccessPoint *ap,
            gpointer user_data)
{
    GSList *items;
    GSList *iter;

    g_signal_handlers_disconnect_by_func (ap, "notify", ap_updated);

    items = nmn_device_handler_get_items (NMN_DEVICE_HANDLER (user_data));
    for (iter = items; iter; iter = iter->next) {
        NmnWifiItem *item = NMN_WIFI_ITEM (iter->data);
        NMAccessPoint *new_ap;

        /* not us */
        if (nmn_wifi_item_get_ap (item) != ap)
            continue;

        /* Do we have another compatible AP for this item? */
        new_ap = find_best_compatible_ap (nm_device_wifi_get_access_points (device), ap);
        if (new_ap)
            nmn_wifi_item_set_ap (item, new_ap);
        else
            /* Nope, remove the item */
            nmn_item_remove_requested (NMN_ITEM (item));
    }
}

static void
connection_added (NmnDeviceHandler *handler,
                  NMSettingsConnectionInterface *connection)
{
    NMSettingConnection *s_con;
    const char *connection_type;
    NMDeviceWifi *device;
    NMAccessPoint *ap;
    GtkWidget *item;

    s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (NM_CONNECTION (connection), NM_TYPE_SETTING_CONNECTION));
    connection_type = nm_setting_connection_get_connection_type (s_con);

    if (!connection_type || strcmp (connection_type, NM_SETTING_WIRELESS_SETTING_NAME))
        /* Not a wifi connection */
        return;

    /* Make sure it doesn't exist yet */
    if (nmn_device_handler_get_item_for_connection (handler, connection))
        return;

    device = NM_DEVICE_WIFI (nmn_device_handler_get_device (handler));
    ap = find_best_ap_for_connection (device, NM_CONNECTION (connection));
    if (ap) {
        item = nmn_wifi_item_new (nmn_device_handler_get_nm_data (handler), device, ap);
        g_object_set (item, NMN_NETWORK_ITEM_CONNECTION, NM_CONNECTION (connection), NULL);
        nmn_network_item_set_delete_visible (NMN_NETWORK_ITEM (item), TRUE);
        nmn_device_handler_add_item (handler, NMN_NETWORK_ITEM (item));
    }
}

static void
ap_added (NMDeviceWifi *device,
          NMAccessPoint *ap,
          gpointer user_data)
{
    NmnDeviceHandler *handler = NMN_DEVICE_HANDLER (user_data);
    GSList *list;
    GSList *iter;

    /* Catch the signals of the new AP */
    g_signal_connect (ap, "notify", G_CALLBACK (ap_updated), user_data);
    ap_updated (ap, NULL, user_data);

    /* Maybe there's an existing connection for it which hasn't been added yet? */
    list = nmn_device_handler_get_connections (handler);
    for (iter = list; iter; iter = iter->next)
        connection_added (handler, NM_SETTINGS_CONNECTION_INTERFACE (iter->data));

    g_slist_free (list);
}

static gboolean
active_ap_changed_cb (gpointer user_data)
{
    NmnDeviceHandler *handler = NMN_DEVICE_HANDLER (user_data);
    NMDeviceWifi *device;
    GSList *items;
    GSList *iter;
    NMAccessPoint *active_ap;

    device = NM_DEVICE_WIFI (nmn_device_handler_get_device (handler));
    active_ap = nm_device_wifi_get_active_access_point (device);
    items = nmn_device_handler_get_items (NMN_DEVICE_HANDLER (user_data));

    for (iter = items; iter; iter = iter->next) {
        NmnWifiItem *item = NMN_WIFI_ITEM (iter->data);

        /* Don't touch inactive items */
        if (nmn_network_item_get_status (NMN_NETWORK_ITEM (item)) == NMN_NETWORK_ITEM_STATUS_DISCONNECTED)
            continue;

        /* If an item gets deactivated, find a new best AP for it */
        if (!active_ap)
            active_ap = find_best_compatible_ap (nm_device_wifi_get_access_points (device),
                                                 nmn_wifi_item_get_ap (item));

        if (active_ap && nmn_wifi_item_get_ap (item) != active_ap)
            nmn_wifi_item_set_ap (item, active_ap);
    }

    return FALSE;
}

static void
active_ap_changed (NMDeviceWifi *device,
                   GParamSpec *pspec,
                   gpointer user_data)
{
    /* We need to delay this a bit, otherwise we haven't received all the other state
       change notification yet and we don't know the correct state */
    g_timeout_add_seconds (2, active_ap_changed_cb, user_data);
}

static void
wifi_toggled (NmnNMData *nm_data,
              gboolean active,
              gpointer user_data)
{
    NmnDeviceHandler *handler = NMN_DEVICE_HANDLER (user_data);

    if (active)
        nmn_device_handler_add_items (handler);
    else
        nmn_device_handler_remove_items (handler);
}

static GObject*
constructor (GType type,
             guint n_construct_params,
             GObjectConstructParam *construct_params)
{
    GObject *object;
    NmnNMData *nm_data;
    NmnWifiHandlerPrivate *priv;
    NMDeviceWifi *device;
    const GPtrArray *aps;
    int i;

    object = G_OBJECT_CLASS (nmn_wifi_handler_parent_class)->constructor
        (type, n_construct_params, construct_params);

    if (!object)
        return NULL;

    priv = GET_PRIVATE (object);

    nm_data = nmn_device_handler_get_nm_data (NMN_DEVICE_HANDLER (object));
    priv->wifi_toggled_id = g_signal_connect (nm_data, "wifi-toggled", G_CALLBACK (wifi_toggled), object);

    device = NM_DEVICE_WIFI (nmn_device_handler_get_device (NMN_DEVICE_HANDLER (object)));

    priv->ap_added_id   = g_signal_connect (device, "access-point-added", G_CALLBACK (ap_added), object);
    priv->ap_removed_id = g_signal_connect (device, "access-point-removed", G_CALLBACK (ap_removed), object);
    priv->ap_changed_id = g_signal_connect (device, "notify::" NM_DEVICE_WIFI_ACTIVE_ACCESS_POINT,
                                            G_CALLBACK (active_ap_changed), object);

    aps = nm_device_wifi_get_access_points (device);
    for (i = 0; aps && i < aps->len; i++) {
        NMAccessPoint *ap = NM_ACCESS_POINT (g_ptr_array_index (aps, i));

        g_signal_connect (ap, "notify", G_CALLBACK (ap_updated), object);
    }

    return object;
}

static void
nmn_wifi_handler_init (NmnWifiHandler *handler)
{
}

static void
dispose (GObject *object)
{
    NmnWifiHandlerPrivate *priv = GET_PRIVATE (object);
    NmnNMData *nm_data;
    NMDeviceWifi *device;
    const GPtrArray *aps;
    int i;

    if (priv->disposed)
        return;

    nm_data = nmn_device_handler_get_nm_data (NMN_DEVICE_HANDLER (object));
    g_signal_handler_disconnect (nm_data, priv->wifi_toggled_id);

    device = NM_DEVICE_WIFI (nmn_device_handler_get_device (NMN_DEVICE_HANDLER (object)));

    aps = nm_device_wifi_get_access_points (device);
    for (i = 0; aps && i < aps->len; i++) {
        NMAccessPoint *ap = NM_ACCESS_POINT (g_ptr_array_index (aps, i));

        g_signal_handlers_disconnect_by_func (ap, "notify", ap_updated);
    }

    g_signal_handler_disconnect (device, priv->ap_added_id);
    g_signal_handler_disconnect (device, priv->ap_removed_id);
    g_signal_handler_disconnect (device, priv->ap_changed_id);

    priv->disposed = TRUE;

    G_OBJECT_CLASS (nmn_wifi_handler_parent_class)->dispose (object);
}

static void
nmn_wifi_handler_class_init (NmnWifiHandlerClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    NmnDeviceHandlerClass *handler_class = NMN_DEVICE_HANDLER_CLASS (class);

    g_type_class_add_private (object_class, sizeof (NmnWifiHandlerPrivate));

    object_class->constructor = constructor;
    object_class->dispose = dispose;

    handler_class->connection_added = connection_added;
}
